from cython cimport size_t
from libcpp.string cimport string
from libcpp.unordered_map cimport unordered_map

cdef class WordPieceProcessor:
    cdef unordered_map[string, size_t] initial_pieces
    cdef unordered_map[string, size_t] continuation_pieces



from typing import List, Tuple
from cython.operator cimport dereference as deref
from libcpp cimport pair

cdef struct PieceMatch:
    bint found
    int prefix_len
    int piece_id

cdef class WordPieceProcessor:
    def __init__(self, pieces: List[str]):
        for idx, piece in enumerate(pieces):
            if piece.startswith("##"):
                self.continuation_pieces[piece[2:].encode('utf8')] = idx
            else:
                self.initial_pieces[piece.encode('utf8')] = idx

    def encode(self, token: str) -> Tuple[List[int], List[str]]:
        """
        Encode a token using word pieces. The piece identifiers are
        returned. If no piece could be found for a suffix, the special
        piece identifier `-1` is used.

            token: The token to encode as pieces.
        """
        token_ids = []
        token_pieces = []
        cdef const unordered_map[string, size_t]* pieces = &self.initial_pieces
        prefix = ""
        cdef PieceMatch match
        while token:
            match = _find_longest_prefix(pieces, token)
            if not match.found:
                token_ids.append(-1)
                token_pieces.append(None)
                return token_ids, token_pieces
            token_ids.append(match.piece_id)
            token_pieces.append(f"{prefix}{token[:match.prefix_len]}")
            token = token[match.prefix_len:]
            pieces = &self.continuation_pieces
            prefix = "##"
        return token_ids, token_pieces

    def get_initial(self, piece: str) -> int:
        cdef unordered_map[string, size_t].const_iterator iter
        iter = self.initial_pieces.const_find(piece.encode('utf8'))
        if iter == self.initial_pieces.end():
            raise KeyError(f"unknown initial piece {piece}")
        else:
            return deref(iter).second


    @staticmethod
    def from_file(filename: str) -> WordPieceProcessor:
        with open(filename, encoding="utf8") as f:
            pieces = [line.strip() for line in f]
        return WordPieceProcessor(pieces)

    def to_list(self):
        pieces = [None] * (self.initial_pieces.size() + self.continuation_pieces.size())

        for piece, idx in self.initial_pieces:
            pieces[idx] = piece.decode('utf8')

        for piece, idx in self.continuation_pieces:
            pieces[idx] = f"##{piece.decode('utf8')}"

        return pieces


cdef PieceMatch _find_longest_prefix(const unordered_map[string, size_t]* pieces, str token):
    cdef PieceMatch match
    cdef unordered_map[string, size_t].const_iterator idx
    match.found = False

    for prefix_len in reversed(range(len(token) + 1)): 
        idx = pieces.const_find(token[:prefix_len].encode("utf8"))
        if idx != pieces.end():
            match.found = True
            match.prefix_len = prefix_len
            match.piece_id = deref(idx).second
            return match

    return match



from cython.operator cimport dereference as deref

cdef class SentencePieceProcessor:
    def __cinit__(self):
        self.spp.reset(new CSentencePieceProcessor())

    def __init__(self):
        pass

    def __len__(self):
        _check_status(deref(self.spp).status())
        return deref(self.spp).GetPieceSize()

    @staticmethod
    def from_file(str filename):
        cdef SentencePieceProcessor processor = SentencePieceProcessor.__new__(SentencePieceProcessor)
        _check_status(deref(processor.spp).Load(filename.encode("utf-8")))
        return processor

    @staticmethod
    def from_protobuf(bytes protobuf):
        cdef SentencePieceProcessor processor = SentencePieceProcessor.__new__(SentencePieceProcessor)
        cdef string_view protobuf_view = string_view(protobuf, len(protobuf))
        _check_status(deref(processor.spp).LoadFromSerializedProto(protobuf_view))
        return processor

    def to_protobuf(self):
        cdef string serialized = deref(self.spp).serialized_model_proto()
        return bytes(serialized)

    def decode_from_ids(self, list ids):
        cdef vector[int] c_ids
        for piece_id in ids:
            if not isinstance(piece_id, int):
                raise TypeError("Pieces must be of type `int` when decoding from ids")
            c_ids.push_back(piece_id)
        cdef string output
        _check_status(deref(self.spp).Decode(c_ids, &output))
        return output.decode("utf-8")

    def decode_from_pieces(self, list pieces):
        cdef vector[string] c_strings
        for piece in pieces:
            if not isinstance(piece, str):
                raise TypeError("Pieces must be of type `str` when decoding from pieces")
            c_strings.push_back(piece.encode("utf-8"))
        cdef string output
        _check_status(deref(self.spp).Decode(c_strings, &output))
        return output.decode("utf-8")

    def encode(self, str sentence):
        cdef SentencePieceText text = self._encode(sentence)

        cdef int idx
        cdef SentencePieceText_SentencePiece piece
        pieces = []
        ids = []
        for idx in range(text.pieces_size()):
            piece = text.pieces(idx)
            ids.append(piece.id())
            pieces.append(piece.piece().decode("utf-8"))

        return ids, pieces

    def encode_as_ids(self, str sentence):
        cdef SentencePieceText text = self._encode(sentence)

        cdef int idx
        ids = []
        for idx in range(text.pieces_size()):
            ids.append(text.pieces(idx).id())

        return ids

    def encode_as_pieces(self, str sentence):
        cdef SentencePieceText text = self._encode(sentence)

        cdef int idx
        cdef SentencePieceText_SentencePiece piece
        pieces = []
        for idx in range(text.pieces_size()):
            piece = text.pieces(idx)
            pieces.append(text.pieces(idx).piece().decode("utf-8"))

        return pieces

    def bos_id(self):
        _check_status(deref(self.spp).status())
        return deref(self.spp).bos_id()

    def eos_id(self):
        _check_status(deref(self.spp).status())
        return deref(self.spp).eos_id()

    def pad_id(self):
        _check_status(deref(self.spp).status())
        return deref(self.spp).pad_id()

    def unk_id(self):
        _check_status(deref(self.spp).status())
        return deref(self.spp).unk_id()

    cdef SentencePieceText _encode(self, str sentence) except *:
        sentence_bytes = sentence.encode("utf-8")
        cdef SentencePieceText text;
        _check_status(deref(self.spp).Encode(sentence.encode("utf-8"), &text))
        return text

cdef _check_status(Status status):
    cdef code = <int> status.code()
    if code == <int> kOk:
        return
    elif code == <int> kNotFound:
        raise OSError(status.error_message().decode("utf-8"))
    elif code == <int> kInternal:
        raise RuntimeError(status.error_message().decode("utf-8"))
    else:
        raise ValueError(status.error_message().decode("utf-8"))



from cutlery import WordPieceProcessor
from pathlib import Path
import pytest


@pytest.fixture(scope="module")
def test_dir(request):
    return Path(request.fspath).parent


@pytest.fixture
def toy_processor_from_file(test_dir):
    return WordPieceProcessor.from_file(str(test_dir / "toy-word-pieces.txt"))


@pytest.fixture
def toy_processor():
    return WordPieceProcessor(["voor", "##tie", "coördina", "##kom", "##en"])


def test_word_piece_processor_small(toy_processor):
    assert toy_processor.encode("voor") == ([0], ["voor"])
    assert toy_processor.encode("voorman") == ([0, -1], ["voor", None])
    assert toy_processor.encode("coördinatie") == ([2, 1], ["coördina", "##tie"])
    assert toy_processor.encode("voorkomen") == ([0, 3, 4], ["voor", "##kom", "##en"])


def test_to_list(toy_processor):
    assert toy_processor.to_list() == ["voor", "##tie", "coördina", "##kom", "##en"]


def test_from_file(toy_processor_from_file):
    assert toy_processor_from_file.to_list() == [
        "voor",
        "##tie",
        "coördina",
        "##kom",
        "##en",
    ]
